#!/usr/bin/env ruby
# -*- coding: utf-8 -*-

require './lib/rdparse.rb'

class FuncDef
  attr_accessor :type, :identifier, :param_list, :stmt_list
  def initialize(type, identifier, param_list, stmt_list)
    @type, @identifier, @param_list, @stmt_list = 
      [type, identifier, param_list, stmt_list]
  end

  def to_a
    retarr = []
    retarr << identifier
    retarr << type
    if param_list.length > 0 then
      retarr << param_list
    else
      retarr << []
    end
    stmt_list.each do |line|
      retarr << line
    end
    retarr
  end
end

FunctionParam = Struct.new(:type, :identifier)
DeclStmt = Struct.new(:identifier, :type)
AssignStmt = Struct.new(:identifier, :type, :value)
ReassignStmt = Struct.new(:identifier, :expression)

ReadStmt = Struct.new(:source, :target)
PrintStmt = Struct.new(:source)
PrintLnStmt = Struct.new(:source)
PrintfStmt = Struct.new(:source, :target)

ReturnStmt = Struct.new(:expression)
BreakStmt = Struct.new(:expression)

WhileStmt = Struct.new(:condition, :statements)
ForStmt = Struct.new(:identifier, :condition, :inc_value, :statements)

AritmExpr = Struct.new(:operator, :operand, :operand1)
NpredExpr = Struct.new(:operator, :operand)
PredExpr = Struct.new(:operator, :operand, :operand1)
RelPredExpr = Struct.new(:operator, :operand, :operand1)

IfStmt = Struct.new(:condition, :statements, :elsif_list, :else)
Elsif = Struct.new(:condition, :statements)
Else = Struct.new(:statements)

FuncCall = Struct.new(:name, :arguments)

Identifier = Struct.new(:name)

Booltype = Struct.new(:name)

Sstring = Struct.new(:value)
Numeric = Struct.new(:value)

class LangParser

  def initialize
    @langParser = Parser.new("Language Parser") do

      token(/^#.*$/)
      token(/\n/)
      token(/\\n/)
      token(/".*"/) {|s| s}
      token(/\s+/)
      token(/(\-\d+|\d+)\.\d+/) {|d| d.to_f}
      token(/(\-\d+|\d+)/) {|d| d.to_i}
      token(/[\w+\.*]+/){|m| m}
      token(/\w+/){|m| m}
      token(/\-\>/) {|m| m}
      token(/(\<\=|\>\=)/) {|m| m}
      token(/./) {|m| m}

      start :program do
        match(:program, :func_def) {
          |func_lst, func_def| 
          func_lst << func_def
          func_lst
        }
        match(:func_def){ 
          |func_def| 
          [func_def]
        }
      end

      rule :func_def do
        match('defun', :type, :identifier, '(', :param_list, ')', '->', 
              :stmt_list, :terminator) { 
          | _, type, identifier, _, param_list, _, _, stmt_list, _ | 
          FuncDef.new(type, identifier, param_list, stmt_list)
        }
        match('defun', :identifier, '(', :param_list, ')', '->', 
              :stmt_list, :terminator) {
          | _, identifier, _, param_list, _, _, stmt_list, _ |
          FuncDef.new('void', identifier, param_list, stmt_list)
        }
        match('defun', :type, :identifier, '(', ')', '->', :stmt_list, 
              :terminator) { 
          | _, type, identifier, _, _, _, stmt_list, _ | 
          FuncDef.new(type, identifier, [], stmt_list)
        }
        match('defun', :identifier, '(', ')', '->', :stmt_list, 
              :terminator) {
          | _, identifier, _, _, _, stmt_list, _|
          FuncDef.new('void', identifier, [], stmt_list)
        }
        match('defun', :type, :identifier, '->', :stmt_list, 
              :terminator) { 
          | _, type, identifier, _, stmt_list, _ | 
          FuncDef.new(type, identifier,[],stmt_list)
        }
        match('defun', :identifier, '->', :stmt_list, :terminator) { 
          | _, identifier, _, stmt_list, _ | 
          FuncDef.new('void', identifier, [], stmt_list)
        }
      end

      rule :stmt_list do
        match(:stmt_list, :stmt, :terminator) { 
          |stmt_list, stmt, _| 
          stmt_list << stmt
        }
        match(:stmt, :terminator) {
          |stmt,_| 
          [stmt]
        }
      end

      rule :stmt do
        match(:func_def) {
          |stmt|
          stmt
        }
        match(:io_stmt) {
          |stmt|
          stmt
        }
        match(:func_call) {
          |call| 
          call
        } 
        match(:iter_stmt) { |stmt|
          stmt
        }
        match(:assign_stmt) {
          |stmt| 
          stmt
        }
        match(:decl_stmt) { 
          |stmt| 
          stmt
        }
        match(:selection_stmt) { 
          |stmt| 
          stmt 
        }
        match(:return_stmt) { 
          |stmt| 
          stmt
        }
        match(:break_stmt) {
          |stmt|
          stmt
        }
      end

      rule :io_stmt do
        match(:print_stmt) { 
          |stmt| 
          stmt
        }
        match(:read_stmt) { 
          |stmt| 
          stmt
        }
      end

      rule :read_stmt do
        match('read', '(', 'stdin', ',', :identifier, ')'){ 
          |_,_,_,_,identifier,_|
          ReadStmt.new('stdin', identifier)
        }
        match('read', '(', :string, ',', :identifier, ')'){ 
          |_,_,source,_,target,_|
          ReadStmt.new(source,target)
        } 
      end

      rule :print_stmt do
        match('println', :expression) { 
          |_, expr| 
          PrintLnStmt.new(expr)
        }
        match('println') {
          |_| 
          PrintLnStmt.new(Sstring.new(""))
        }
        match('print', :expression){ 
          |_,expr| 
          PrintStmt.new(expr)
        }
        match('printf', '(', :expression, ',',  :string, ')'){ 
          |_, _, expression, _, filename, _ | 
          PrintfStmt.new(expression, filename)
        }
      end

      rule :decl_stmt do
        match(:identifier, :type){
          | identifier, type |
          DeclStmt.new(identifier, type)
        }
      end

      rule :assign_stmt do
        match(:identifier, :type, ':', :expression) { 
          | identifier, type, _, expression | 
          AssignStmt.new(identifier, type, expression)
        }
        match(:identifier, ':', :expression) {
          | identifier, _ , expression |
          ReassignStmt.new(identifier, expression)
        }
      end

      rule :iter_stmt do
        match(:while_stmt){|stmt| stmt}
        match(:for_stmt){|stmt| stmt}
      end

      rule :while_stmt do
        match('while', '(', :pred_expr, ')', '->', :stmt_list){ 
          |_,_,expr,_,_,stmts|
          WhileStmt.new(expr, stmts)
        }
      end

      rule :for_stmt do
        match('for', :identifier, ',', '(', :pred_expr, ')', ',', 
              :numeric, '->', :stmt_list){ 
          |_,identifier,_,_,expr,_,_,inc,_,stmts|
          ForStmt.new(identifier, expr, inc, stmts)
        }
        match('for', :identifier, ',', '(', :pred_expr, ')', ',', 
              :identifier, '->', :stmt_list){ 
          |_,identifier,_,_,expr,_,_,inc,_,stmts|
          ForStmt.new(identifier, expr, inc, stmts)
        }
        match('for', :identifier, ',', '(', :pred_expr, ')', '->', 
              :stmt_list){ 
          |_,identifier,_,_,expr,_,_,stmts|
          ForStmt.new(identifier, expr, Numeric.new(1), stmts)
        }
      end

      rule :selection_stmt do
        match(:if_stmt){|stmt| stmt}
      end

      rule :if_stmt do
        match('if', '(', :pred_expr, ')', '->', :stmt_list, :elsif_list, 
              :else){ 
          |_,_,expr,_,_,stmts,elsif_list,els|
          IfStmt.new(expr,stmts,elsif_list,els)
        }
        match('if', '(', :pred_expr, ')', '->', :stmt_list, :elsif_list){ 
          |_,_,expr,_,_,stmts,elsif_list|
          IfStmt.new(expr,stmts,elsif_list,nil)
        }
        match('if', '(', :pred_expr, ')', '->', :stmt_list, :else){ 
          |_,_,expr,_,_,stmts,els|
          IfStmt.new(expr,stmts,nil,els)
        }
        match('if', '(', :pred_expr, ')', '->', :stmt_list){ 
          |_,_,expr,_,_,stmts|
          IfStmt.new(expr,stmts,nil,nil)
        }
      end

      rule :elsif_list do
        match(:elsif_list, :elsif) {
          |elsif_list, els| 
          elsif_list << els
        }
        match(:elsif){|els| [els]}
      end

      rule :elsif do
        match('elsif', '(', :pred_expr, ')', '->', :stmt_list){ 
          |_,_,expr,_,_,stmts| 
          Elsif.new(expr,stmts)
        }
      end

      rule :else do
        match('else', '->', :stmt_list){ 
          |_,_,stmts|
          Else.new(stmts)
        }
      end

      rule :return_stmt do
        match('return', :expression){ 
          |_, expr| 
          ReturnStmt.new(expr) 
        }
      end

      rule :break_stmt do
        match('break'){
          |_, _|
          BreakStmt.new("Break")
        }
      end

      rule :expression do
        match('(', :aritm_expr, ')'){ 
          |_,a_expr,_| 
          a_expr 
        }
        match('(', :pred_expr, ')'){ 
          |_,p_expr,_| 
          p_expr 
        }
        match(:func_call){
          |func_call| 
          func_call
        }
        match(:literal){ 
          |literal|
          literal
        }
        match(:identifier){ 
          |identifier| 
          identifier
        }
      end

      rule :aritm_expr do
        match(/(\+|\-|\*|\/|\^|\%)/, :expression, :expression){ 
          |op,exp,exp1| 
          AritmExpr.new(op, exp, exp1)
        }
      end

      rule :pred_expr do
        match('!', :expression){ 
          |oper, expression| 
          NpredExpr.new(oper, expression)
        }
        match(:rel_oper, :expression, :expression){
          |oper, expr, expr1| 
          RelPredExpr.new(oper, expr, expr1)
        }
        match(:log_oper, :expression, :expression){
          |oper, expr, expr1| 
          PredExpr.new(oper, expr, expr1)
        }
      end

      rule :log_oper do
        match(/(\!|\&|\|)/){
          |oper| 
          oper
        }
      end

      rule :rel_oper do
        match(/(\<|\>|\<\=|\>\=|\=)/){
          |oper| 
          oper
        }
      end

      rule :func_call do
        match(:identifier, '(', :arg_list, ')') { 
          | identifier, _, args, _ |
          FuncCall.new(identifier, args)
        }
        match(:identifier, '(', ')') {
          | identifier, _, _|
          FuncCall.new(identifier,[])
        }
      end

      rule :param_list do
        match(:param_list, ',', :parameter) { 
          |lst, _, parameter| 
          lst << parameter
          
        }
        match(:parameter) {
          |param|
          [param]
        }
      end

      rule :parameter do
        match(:type, :identifier) { 
          |type, identifier|
          FunctionParam.new(type, identifier)
        }
      end

      rule :arg_list do
        match(:arg_list, ',', :expression) {
          |lst, _, expression| 
          lst << expression
        }
        match(:expression){
          |expression| 
          [expression]
        }
      end

      rule :literal do
        match(:numeric) {
          |num| 
          num
        }
        match(:booltype) {
          |booltype| 
          booltype
        }
        match(:string) {
          |str|
          str
        }
      end

      rule :numeric do
        match(:float) {
          |f| 
          Numeric.new(f)
        }
        match(Fixnum) {
          |i| 
          Numeric.new(i)
        }
      end

      rule :type do
        match('float') {
          |f| 
          'Float'
        }
        match('int') {
          |i| 
          'Fixnum'
        }
        match('string') {
          |s| 
          'String'
        }
        match('void') {
          |t| 
          t
        }
        match('bool') {
          |b| 
          b
        }
      end
      
      rule :identifier do
        match(/[a-zA-Z]+[a-zA-Z_]*/) {
          |i| 
          Identifier.new(i)
        }
      end

      rule :float do
        match(Float) {
          |f| 
          f
        }
      end

      rule :booltype do
        match(/(true|false)/) { 
          |type| 
          Booltype.new(type) 
        }
      end

      rule :string do
        match(/\".*\"/){
          |s| 
          Sstring.new(s.to_s.gsub(/\"/,''))
        }
      end
      
      rule :terminator do
        match(';') {
          |t| 
          t
        }
      end
    end
  end

  def parse str
    @langParser.parse str
  end

  def log(state = true)
    @langParser.logger.level = state ? Logger::DEBUG : Logger::WARN
  end

end

def parse_file filename
  lp = LangParser.new
  lp.log false
  f = File.open(filename)
  functions = lp.parse(f.read)
  f.close
  functions
end

def parse_string string
  lp = LangParser.new
  lp.log false
  functions = lp.parse string
  functions
end

