#!/usr/bin/env ruby
# -*- coding: utf-8 -*-

=begin
An interpreter for the RIPPLE language.
=end

#Just to get rid of a pesky little warning..
$VERBOSE = nil

require './library/parser.rb'
require './library/math_logic.rb'
require './library/function_expression_handler.rb'


LANG_NAME = "RIPPLE"

#Set to true if debugging should be enabled
$debug = false

STDOUT.sync = true
STDIN.sync = true

#The statement types that should be scopeable
$scopeable = [:WhileStmt, :ForStmt, :IfStmt]

##
# This class can be seen as the container
# for all functions and it is this class
# that handle all interactions between
# different functions defined at the 
# top level.
##
class Program
  attr_accessor :functions

  def initialize
    @functions = {}
  end

  def get_function(function_name)
    puts "in get_function container" if $debug
    if function_name.class == Array then
      function_name = function_name.join()
    end
    print "#{function_name}\n" if $debug
    if(functions.key?(function_name))
      Function.new(functions[function_name],self)
    end
  end

  def has_function(function_name)
    puts "in has function (container)" if $debug
    print "#{function_name}\n" if $debug
    puts function_name.class if $debug
    if function_name.class == Array then
      puts "function_name.class == Array" if $debug
      function_name = function_name.join()
    end
    print "got function_name: #{function_name}\n" if $debug
    return (functions[function_name] != nil)
  end

  def add_function(function)
    if has_function(function.identifier.name) then
      raise Exception, "Duplicate functions #{function.identifier.name}!"
    end
    functions[function.identifier.name] = function
  end

  def get_function_parameter_list(function_name)
    functions[function_name].param_list
  end

  def print_functions
    functions.each_key do |key|
      puts key
    end
  end

  def run
    get_function("init").call
  end

end

##
# This class represents a Function. This is
# what a program in RIPPLE is made up of.
##
class Function
  
  attr_accessor :identifier, :parent, :return_type, 
  :parameter_list, :variables, :statements, :functions, :cm,
  :variable_types

  def initialize(function, parent)
    @identifier = function.identifier.name
    @parent = parent
    @name = function.identifier
    @variables = Hash.new
    @variable_types = Hash.new
    @cm = CommonMethods.new(self)

    if($debug) then
      puts function.param_list.class
      function.param_list.each do |param|
        puts param.class
      end
    end
    @parameter_list = function.param_list

    if(function.type != 'void') then
      variables["_return_value_"] = nil
    end
    @statements = function.stmt_list
    @functions = {}
  end

  def print_statements
    statements.each do |statement|
      puts statement
    end
  end
  
  # This is the method that is called when
  # the program wants to use the function
  # and this is where all work is done.
  def call(*args)
    if($debug) then
      print_variables
      puts "In function #{identifier}"
      puts "args: "
      args.each do |arg|
        print "#{arg},"
      end
      puts "statements:"
      print_statements
      print "\n"
    end

    if (args.length == 1) then
      args = args[0]
    end

    if (args.length != parameter_list.length)
      raise Exception, "Argument list length doesn't match" + 
        "parameter list length in call to function #{identifier}"
    end

    # Iterate through the parameter list
    # to make sure that the provided
    # arguments match.
    parameter_list.each_index do |idx|
      if (parameter_list[idx].type == "bool") then
        if(args[idx].class.to_s != "FalseClass" && 
           args[idx].class.to_s != "TrueClass") then
          raise Exception, "Argument #{idx} not valid for function " + 
            "#{identifier}. Argument should be " + 
            "\"#{parameter_list[x].type}\"" +
            ", was\ #{args[x].class}"
        end
      elsif (parameter_list[idx].type != args[idx].class.to_s)
        raise Exception, "Argument #{x} not valid for function " + 
          "#{identifier}. Argument should be " + 
          "\"#{parameter_list[x].type}\", was #{args[x].class}"
      end
    end

    parameter_list.each_index do |idx|
      variables[parameter_list[idx].identifier.name] = args[idx]
    end

    # Iterate through the statements
    # in the function and evaluate them.
    statements.each do |statement|
      if statement.class == ReturnStmt then
        evaluate(statement, self)
        break
      elsif $scopeable.index("#{statement.class}".to_sym) != nil then
        Scope.new(self, statement).run
      else
        puts "Should do something" if $debug
        evaluate(statement, self)
      end
    end

    print_variables if $debug

    # If the function was defined with
    # a return type then return the value
    # that should be stored at "_return_value_"
    if(variables.key?("_return_value_")) then
      return variables["_return_value_"]
    end
  end

  def method_missing(sym, *args)
    puts "In method missing function, #{sym}" if $debug
    instance_eval("cm.#{sym}(#{args})")
  end

end

##
# A Scope in RIPPLE is the "lowest"
# container type. It simply has one
# statement which it runs. The 
# statement on the other hand can 
# itself have many statements.
##
class Scope
  
  attr_accessor :parent, :statement, :variables, :cm, :functions,
  :variable_types

  def initialize(parent, statement)
    @parent = parent
    @statement = statement
    @variables = Hash.new
    @variable_types = Hash.new
    @functions = Hash.new
    @cm = CommonMethods.new(self)
  end

  def run
    puts "In run" if $debug
    evaluate(statement, self)
    print_variables if $debug
  end

  def method_missing(sym, *args)
    puts "In method missing scope" if $debug
    self.instance_eval("cm.#{sym}(#{args})")
  end

end

##
# This class represents common methods
# that the classes Function and Scope
# might want to use. Every method in 
# this class operates relative to the
# context specified when initializing.
##
class CommonMethods
  
  attr_accessor :context

  # Initialize with the callers instance
  def initialize(context)
    @context = context
  end

  def create_variable(args)
    name, value = args[0], args[1]
    puts "In create variable" if $debug
    if (context.variables.key?(name)) then
      raise Exception, "Cannot declare the same variable " + 
        "(\"#{name}\") multiple times!"
    else
      value = context.get_correct_value(value, context.variable_types[name])
      context.variables[name] = value
    end
  end

  def get_correct_value(args)
    value, type = args[0], args[1]
    
    if(type == "Int")
      value = value.to_i
    elsif(type == "Float")
      value = value.to_f
    elsif(type == "String")
      value = value.to_s
    elsif(type == "Bool")
      if value == "true"
        value = true
      else
        value = false
      end
    end
    return value
  end

  def set_variable_type(args)
    name, type = args[0], args[1]
    context.variable_types[name] = type
  end

  def set_variable(args)
    name, value = args[0], args[1]
    if (context.variables.key?(name)) then
      type = context.variable_types[name]
      value = context.get_correct_value(value, type)
      context.variables[name] = value
    else
      context.parent.set_variable(name, value)
    end
  end

  def unset_variable(args)
    name = args[0]
    context.variables.delete_if { |key,val| 
      key == name
    }
  end

  def get_variable(args)
    puts "in get variable" if $debug
    puts "args = #{args}" if $debug
    name = args
    if args.class == Array then
      name = args.join()
    end
    if (context.variables.key?(name)) then
      if (context.variables[name] == nil) then
        raise Exception, "Variable #{name} uninitialized!" + 
          " Cannot be used this way"
      else
        context.variables[name]
      end
    else
      puts "Did not find variable #{name}" if $debug
      context.parent.get_variable(name)
    end
  end

  def print_variables(args)
    puts "Variables at end of #{context.class}" if $debug
    context.variables.each_pair do |key, val|
      puts "#{key} = #{val}"
    end
  end

  def has_function(function_name)
    if (function_name.class == Array) then
      function_name = function_name.join()
    end
    if (context.functions[function_name] != nil) then
      return true
    else
      return context.parent.has_function(function_name)
    end
  end

  def get_function(function_name)
    if (function_name.class == Array) then
      function_name = function_name.join()
    end
    if (context.functions[function_name] != nil) then
      return Function.new(context.functions[function_name], context)
    else
      return context.parent.get_function(function_name)
    end
  end

  # This is the real work horse of this class.
  # It simply takes an argument and decides
  # what to do with it.
  def eval(arg)
    puts "In eval" if $debug
    if arg.class == Array
      arg = arg.join()
    end
    instance_eval(arg)
  end

  def method_missing(sym, *args)
    self.instance_eval("context.parent.#{sym}(#{args})")
  end

end


##
# Main evaluation loop. This determines
# what kind of statement that is passed
# to the method and takes action according
# to that.
#
# All calls made in this method is to
# methods defined in:
# function_expression_handler.rb
##
def evaluate(stmt, context)

  if stmt.class == PrintStmt
    result = evaluate(stmt.source, context)
    context.eval("print \"#{result}\"")
  elsif stmt.class == PrintLnStmt
    result = evaluate(stmt.source, context)
    context.eval("puts \"#{result}\"")
  elsif stmt.class == AssignStmt
    context.set_variable_type(stmt.identifier.name, stmt.type)
    context.create_variable(stmt.identifier.name, 
                            get_expression(stmt.value, context))
  elsif stmt.class == ReassignStmt
    context.set_variable(stmt.identifier.name,
                         get_expression(stmt.expression, context))
  elsif stmt.class == DeclStmt
    context.set_variable_type(stmt.identifier.name, stmt.type)
    context.create_variable(stmt.identifier.name, nil)
  elsif stmt.class == FuncCall
    handle_FuncCall(stmt, context)
  elsif stmt.class == FuncDef
    context.functions[stmt.identifier.name] = stmt
  elsif stmt.class == ReturnStmt
    context.set_variable("_return_value_", 
                         get_expression(stmt.expression, context))
  elsif stmt.class == ForStmt
    handle_ForStmt(stmt, context)
  elsif stmt.class == WhileStmt
    handle_WhileStmt(stmt, context)
  elsif stmt.class == IfStmt
    handle_IfStmt(stmt, context)
  elsif stmt.class == ReadStmt
    handle_ReadStmt(stmt, context)
  elsif stmt.class == PrintfStmt
    handle_PrintfStmt(stmt, context)
  elsif stmt.class == PredExpr
    get_expression(stmt, context)
  elsif stmt.class == RelPredExpr
    get_expression(stmt, context)
  elsif stmt.class == NpredExpr
    get_expression(stmt, context)
  elsif stmt.class == AritmExpr
    get_expression(stmt, context)
  elsif stmt.class == Identifier
    context.get_variable("#{stmt.name}")
  elsif stmt.class == Sstring
    stmt.value
  elsif stmt.class == Numeric
    stmt.value
  elsif stmt.class == BreakStmt
    "Break"
  end

end

##
# The method that actually sets up the 
# interpreting environment and runs the 
# code.
##
def interpreter(instructions)
  if instructions.nil? and instructions.length <= 0 then
    raise Exception, "Got no instructions"
  end

  # Check for init function. No init = no program.
  has_init = false
  instructions.each do |func|
    puts func.identifier.name if $debug
    if func.identifier.name == "init"
      has_init = true
      break
    end
  end

  if !has_init then
    raise Exception, "No init function found!"
  else

    # Add all functions to the program and run the init function.
    program = Program.new
    instructions.each do |function|
      program.add_function(function)
    end
    program.run
  end

end

##
# Display a friendly welcome message upon 
# start of interpreter.
##
def welcome
  print "####\n" +
    "## Welcome to a partly working interpreter " + 
    "for the #{LANG_NAME} language\n" +
    "## If you feel lost please do \"help\"\n" + 
    "####\n\n"
end

##
# Prints a message with help regarding how 
# to use the interactive console.
##
def help
  print "####\n" +
    "## Some general guidance regarding the usage of this interpreter\n" +
    "## Useful commands:\n" +
    "## exit: exits the interpreter\n" +
    "## help: shows this help\n" +
    "## pb: prints the current buffer\n" +
    "## cb: clears the current buffer\n" +
    "## rm <num>: removes the chosen line from the buffer\n" +
    "## md <num> <line>: modifies the chosen line in the buffer\n" +
    "## ins <num> <line>: inserts specified line at the specified" +
    " number in buffer\n" +
    "## run: try to execute the buffer *fingers crossed*\n" +
    "####\n\n"
end

##
# A nice way to display the error message. This 
# will be used for all exceptions that occur 
# inside the interpreter or indeed all files 
# that is used by the interpreter.
##
def display_error(errmsg)
    puts 
    puts "###"
    puts "An error occurred:"
    puts errmsg.message
    puts "###"
    puts
end

##
# The code that is run when the user wants
# to use the interactive version of the 
# interpreter. This basically takes some
# input and determines the action based 
# on content.
##
def interactive_mode(already_initialized=false)

  if(!already_initialized)
    welcome()
    $buffer = []
  end
  
  #Make sure we can rescue if something goes wrong. ie typo...
  begin
    while true do
      print "> "
      command = gets.chomp
      if command == "exit" || command == "quit"
        break
      elsif command == "help"
        help()
      elsif command == "debug"
        $debug = !$debug
        puts "Debug -> #{$debug}"
      elsif command == "pb" #Print buffer
        puts "Contents of buffer:"
        Range.new(0,$buffer.length).each do |linum|
          puts "#{linum}: #{$buffer[linum]}"
        end
      elsif command == "cb" #Clear buffer
        puts "Clearing buffer..."
        $buffer = Array.new
      elsif command =~ /^md\s*\d+.*/ #Modify line of buffer
        $buffer[command.scan(/\d+/)[0].to_i] = 
          command.sub(/md\s*\d+/,'')
      elsif command =~ /^rm\s*\d+$/ #Remove line from buffer
        $buffer.delete_at(command.scan(/\d+/)[0].to_i)
      elsif command =~ /^ins\s*\d+.*/ #Insert line in buffer
        $buffer.insert(command.scan(/\d+/)[0].to_i, 
                       command.sub(/ins\s*\d+\s*/, ''))
      elsif command == "run"
        # Assemble instructions, parse and run
        instructions = $buffer.join("\n")
        instructions = parse_string instructions
        interpreter instructions
      else
        $buffer.push command
      end
    end
  rescue Exception => e
    display_error(e)
    interactive_mode(true)
  end
  
end

##
# Entry point when running the file. If a 
# filename is given it will send the file 
# to the parser and then run the interpreter 
# with the instructions from the parser, 
# otherwise it will start the interactive
# mode.
##
begin
  if ARGV.length == 0 then #Assume full interactive mode
    interactive_mode(false)
  else #Assume filename given
    if !File.exists? ARGV[0].to_s then #Check that file exists
      puts "The specified file \"#{ARGV[0].to_s}\" " + 
        "doesn't seem to exist."
    else
      if ARGV.length == 2
        if ARGV[1] == "true"
          $debug = true
        end
      end
      instructions = parse_file(ARGV[0].to_s)
      interpreter instructions
    end
  end
rescue Exception => e
  display_error(e)
end
